//
//      coded by un
//            --------------
//                     mindshifter.com
//

using System;
using System.Collections.Generic;
using Microsoft.Xna.Framework;
using XI = Microsoft.Xna.Framework.Input;

namespace Mindshifter
{
    /// <summary>
    /// Defines the type of a discrete input
    /// </summary>
    public enum DiscreteInputEventType
    {
        Up,
        Down,
        Repeat,
    }

    /// <summary>
    /// Event handler for discrete input events such as key presses and mouse clicks
    /// </summary>
    /// <param name="time">The current game time at which the event occurred</param>
    public delegate void DiscreteInputEventHandler(GameTime time);

    /// <summary>
    /// Event handler for mouse move events
    /// </summary>
    /// <param name="dx">The delta x value of the mouse input event</param>
    /// <param name="dy">The delta y value of the mouse input event</param>
    /// <param name="time">The current game time at which the event occurred</param>
    public delegate void MouseMoveEventHandler(int dx, int dy, GameTime time);

    /// <summary>
    /// Event handler for mouse wheel scroll events
    /// </summary>
    /// <param name="dw">The delta wheel value of the mouse input event</param>
    /// <param name="time">The current game time at which the event occurred</param>
    public delegate void MouseScrollEventHandler(int dw, GameTime time);

    /// <summary>
    /// ActionMap provides support for mapping input handling functions to specific inputs.
    /// </summary>
    public class ActionMap
    {
        private class KeyMap : Dictionary<XI.Keys, DiscreteInputEventHandler> { }
        private class ButtonMap : Dictionary<MouseButtons, DiscreteInputEventHandler> { }

        private Dictionary<DiscreteInputEventType, KeyMap>      mKeyEvents;
        private Dictionary<DiscreteInputEventType, ButtonMap>   mMouseButtonEvents;

        /// <summary>
        /// Event called when the mouse position changes
        /// </summary>
        public event MouseMoveEventHandler                      MouseMoved;

        /// <summary>
        /// Event called when the mouse scroll wheel value changes
        /// </summary>
        public event MouseScrollEventHandler                    MouseScrolled;

        public ActionMap()
        {
            mKeyEvents = new Dictionary<DiscreteInputEventType, KeyMap>();
            mMouseButtonEvents = new Dictionary<DiscreteInputEventType, ButtonMap>();

            mKeyEvents[DiscreteInputEventType.Up] = new KeyMap();
            mKeyEvents[DiscreteInputEventType.Down] = new KeyMap();
            mKeyEvents[DiscreteInputEventType.Repeat] = new KeyMap();
            mMouseButtonEvents[DiscreteInputEventType.Up] = new ButtonMap();
            mMouseButtonEvents[DiscreteInputEventType.Down] = new ButtonMap();
            mMouseButtonEvents[DiscreteInputEventType.Repeat] = new ButtonMap();
        }

        /// <summary>
        /// Binds the specified discrete input event handler to be called when
        /// the specified key generates the specified input event type
        /// </summary>
        /// <param name="key">The key that will trigger the input handler</param>
        /// <param name="type">The input event type that will trigger the input handler</param>
        /// <param name="handler">The input handler to call when the specified input
        /// event occurrs</param>
        /// <returns>The input event handler previously bound to the specified input event,
        /// or null if no handler was bound</returns>
        public DiscreteInputEventHandler Bind(XI.Keys key, DiscreteInputEventType type, 
                                                DiscreteInputEventHandler handler)
        {
            DiscreteInputEventHandler prev = null;

            if (mKeyEvents.ContainsKey(type) && mKeyEvents[type].ContainsKey(key))
                prev = mKeyEvents[type][key];

            mKeyEvents[type][key] = handler;

            return prev;
        }

        /// <summary>
        /// Unbinds the input event handler from the specified key and event type
        /// combination
        /// </summary>
        /// <param name="key">The key from which to unbind the current input event
        /// handler</param>
        /// <param name="type">The input event type for the specified key</param>
        /// <returns>The event handler previously bound to the specified input event,
        /// or null if no event handler was bound</returns>
        public DiscreteInputEventHandler Unbind(XI.Keys key, DiscreteInputEventType type)
        {
            if (mKeyEvents[type].ContainsKey(key))
            {
                DiscreteInputEventHandler handler = mKeyEvents[type][key];
                mKeyEvents[type].Remove(key);
                return handler;
            }

            return null;
        }

        /// <summary>
        /// Binds the specified discrete input event handler to be called when
        /// the specified mouse button generates the specified input event type
        /// </summary>
        /// <param name="button">The mouse button that will trigger the input handler</param>
        /// <param name="type">The input event type that will trigger the input handler</param>
        /// <param name="handler">The input handler to call when the specified input
        /// event occurrs</param>
        /// <returns>The input event handler previously bound to the specified input event,
        /// or null if no handler was bound</returns>
        public DiscreteInputEventHandler Bind(MouseButtons button, DiscreteInputEventType type, 
                                                DiscreteInputEventHandler handler)
        {
            DiscreteInputEventHandler prev = null;

            if (mMouseButtonEvents.ContainsKey(type) && mMouseButtonEvents[type].ContainsKey(button))
                prev = mMouseButtonEvents[type][button];

            mMouseButtonEvents[type][button] = handler;

            return prev;
        }

        /// <summary>
        /// Unbinds the input event handler from the specified mouse button and event type
        /// combination
        /// </summary>
        /// <param name="button">The mouse button from which to unbind the current input event
        /// handler</param>
        /// <param name="type">The input event type for the specified mouse button</param>
        /// <returns>The event handler previously bound to the specified input event,
        /// or null if no event handler was bound</returns>
        public DiscreteInputEventHandler Unbind(MouseButtons button, DiscreteInputEventType type)
        {
            if (mMouseButtonEvents[type].ContainsKey(button))
            {
                DiscreteInputEventHandler handler = mMouseButtonEvents[type][button];
                mMouseButtonEvents[type].Remove(button);
                return handler;
            }

            return null;
        }

        /// <summary>
        /// Processes the current and previous input states to determine which
        /// input events have occurred, and triggers the corresponding input
        /// event handlers
        /// </summary>
        /// <param name="keyboardCurrent">The current keyboard state</param>
        /// <param name="keyboardPrev">The previous keyboard state</param>
        /// <param name="mouseCurrent">The current mouse state</param>
        /// <param name="mousePrev">The previous mouse state</param>
        /// <param name="time">The current game time</param>
        public void ProcessInput(XI.KeyboardState keyboardCurrent, XI.KeyboardState keyboardPrev,
            XI.MouseState mouseCurrent, XI.MouseState mousePrev, GameTime time)
        {
            foreach (XI.Keys key in mKeyEvents[DiscreteInputEventType.Up].Keys)
                if (keyboardCurrent[key] == XI.KeyState.Up && keyboardPrev[key] == XI.KeyState.Down)
                    mKeyEvents[DiscreteInputEventType.Up][key](time);

            foreach (XI.Keys key in mKeyEvents[DiscreteInputEventType.Down].Keys)
                if (keyboardCurrent[key] == XI.KeyState.Down && keyboardPrev[key] == XI.KeyState.Up)
                    mKeyEvents[DiscreteInputEventType.Down][key](time);

            foreach (XI.Keys key in mKeyEvents[DiscreteInputEventType.Repeat].Keys)
                if (keyboardCurrent[key] == XI.KeyState.Down && keyboardPrev[key] == XI.KeyState.Down)
                    mKeyEvents[DiscreteInputEventType.Repeat][key](time);

            foreach (MouseButtons button in mMouseButtonEvents[DiscreteInputEventType.Up].Keys)
            {
                XI.ButtonState curstate = (button == MouseButtons.Left)
                    ? (mouseCurrent.LeftButton)
                    : (button == MouseButtons.Right)
                        ? (mouseCurrent.RightButton)
                        : (mouseCurrent.MiddleButton);

                XI.ButtonState prevstate = (button == MouseButtons.Left)
                    ? (mousePrev.LeftButton)
                    : (button == MouseButtons.Right)
                        ? (mousePrev.RightButton)
                        : (mousePrev.MiddleButton);

                if (curstate == XI.ButtonState.Released && prevstate == XI.ButtonState.Pressed)
                    mMouseButtonEvents[DiscreteInputEventType.Up][button](time);
            }

            foreach (MouseButtons button in mMouseButtonEvents[DiscreteInputEventType.Down].Keys)
            {
                XI.ButtonState curstate = (button == MouseButtons.Left)
                    ? (mouseCurrent.LeftButton)
                    : (button == MouseButtons.Right)
                        ? (mouseCurrent.RightButton)
                        : (mouseCurrent.MiddleButton);

                XI.ButtonState prevstate = (button == MouseButtons.Left)
                    ? (mousePrev.LeftButton)
                    : (button == MouseButtons.Right)
                        ? (mousePrev.RightButton)
                        : (mousePrev.MiddleButton);

                if (curstate == XI.ButtonState.Pressed && prevstate == XI.ButtonState.Released)
                    mMouseButtonEvents[DiscreteInputEventType.Down][button](time);
            }

            foreach (MouseButtons button in mMouseButtonEvents[DiscreteInputEventType.Repeat].Keys)
            {
                XI.ButtonState curstate = (button == MouseButtons.Left)
                    ? (mouseCurrent.LeftButton)
                    : (button == MouseButtons.Right)
                        ? (mouseCurrent.RightButton)
                        : (mouseCurrent.MiddleButton);

                XI.ButtonState prevstate = (button == MouseButtons.Left)
                    ? (mousePrev.LeftButton)
                    : (button == MouseButtons.Right)
                        ? (mousePrev.RightButton)
                        : (mousePrev.MiddleButton);

                if (curstate == XI.ButtonState.Pressed && prevstate == XI.ButtonState.Pressed)
                    mMouseButtonEvents[DiscreteInputEventType.Repeat][button](time);
            }   

            if(MouseMoved != null)
            {
                int dx = mouseCurrent.X - mousePrev.X;
                int dy = mouseCurrent.Y - mousePrev.Y;

                if ((dx != 0) || (dy != 0))
                    MouseMoved(dx, dy, time);
            }

            if (MouseScrolled != null)
            {
                int dw = mouseCurrent.ScrollWheelValue - mousePrev.ScrollWheelValue;

                if (dw != 0)
                    MouseScrolled(dw, time);
            }
        }
    }
}