//
//      coded by un
//            --------------
//                     mindshifter.com
//

using System;
using System.Collections.Generic;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Input;

namespace Mindshifter
{
    /// <summary>
    /// IInputManager defines an interface to the InputManager class
    /// </summary>
    public interface IInputManager
    {
        /// <summary>
        /// Gets the current keyboard state
        /// </summary>
        KeyboardState Keyboard { get; }

        /// <summary>
        /// Gets the current mouse state
        /// </summary>
        MouseState Mouse { get; }

        /// <summary>
        /// Gets the change in the mouse x position since input
        /// was last updated
        /// </summary>
        int DX { get; }

        /// <summary>
        /// Gets the change in the mouse y position since input 
        /// was last updated
        /// </summary>
        int DY { get; }

        /// <summary>
        /// Gets the change in the mouse scroll wheel position
        /// since input was last updated
        /// </summary>
        int DZ { get; }

        /// <summary>
        /// Gets the x position of the mouse
        /// </summary>
        int X { get; }

        /// <summary>
        /// Gets the y position of the mouse
        /// </summary>
        int Y { get; }

        /// <summary>
        /// Gets the scroll wheel value of the mouse
        /// </summary>
        int Z { get; }

        /// <summary>
        /// Returns true if the specified keyboard key is pressed
        /// </summary>
        /// <param name="key">The key to test for a key pressed state</param>
        /// <returns>True if the specified keyboard key is pressed, false otherwise</returns>
        bool this[Keys key] { get; }

        /// <summary>
        /// Returns true if the specified mouse button is pressed
        /// </summary>
        /// <param name="button">The mouse button to test for a pressed state</param>
        /// <returns>True if the specified mouse button is pressed</returns>
        bool this[MouseButtons button] { get; }

        /// <summary>
        /// Gets the active ActionMap
        /// </summary>
        ActionMap ActionMap { get; }

        /// <summary>
        /// Pushes an ActionMap onto the action map stack
        /// </summary>
        /// <param name="map">The ActionMap to push onto the ActionMap stack</param>
        void PushActionMap(ActionMap map);

        /// <summary>
        /// Pops the topmost ActionMap from the ActionMap stack
        /// </summary>
        /// <returns>The ActionMap that was popped from the ActionMap stack</returns>
        ActionMap PopActionMap();

        /// <summary>
        /// Returns true if the specified keyboard key is pressed
        /// </summary>
        /// <param name="key">The key to test for a key pressed state</param>
        /// <returns>True if the specified keyboard key is pressed, false otherwise</returns>
        bool Key(Keys key);

        /// <summary>
        /// Returns true if a key down event occurred for the specified key
        /// </summary>
        /// <param name="key">The key to test for a key down event</param>
        /// <returns>True if the specified keyboard key triggered a key down
        /// event, false otherwise</returns>
        bool KeyDown(Keys key);

        /// <summary>
        /// Returns true if a key up event occurred for the specified key
        /// </summary>
        /// <param name="key">The key to test for a key up event</param>
        /// <returns>True if the specified keyboard key triggered a key up
        /// event, false otherwise</returns>
        bool KeyUp(Keys key);

        /// <summary>
        /// Returns true if the specified mouse button is pressed
        /// </summary>
        /// <param name="button">The mouse button to test for a button pressed state</param>
        /// <returns>True if the specified mouse button is pressed, false otherwise</returns>
        bool MouseButton(MouseButtons button);

        /// <summary>
        /// Returns true if a mouse down event occurred for the specified
        /// mouse button
        /// </summary>
        /// <param name="button">The mouse button to test for a mouse down
        /// event</param>
        /// <returns>True if the specified mouse button triggered a mouse
        /// down event</returns>
        bool MouseDown(MouseButtons button);

        /// <summary>
        /// Returns true if a mouse up event occurred for the specified
        /// mouse button
        /// </summary>
        /// <param name="button">The mouse button to test for a mouse up
        /// event</param>
        /// <returns>True if the specified mouse button triggered a mouse
        /// up event
        /// </returns>
        bool MouseUp(MouseButtons button);
    }
}
