//
//      coded by un
//            --------------
//                     mindshifter.com
//

using System;
using System.Collections.Generic;
using Microsoft.Xna.Framework;
using XI = Microsoft.Xna.Framework.Input;

namespace Mindshifter
{
    /// <summary>
    /// Defines the buttons of the mouse
    /// </summary>
    public enum MouseButtons
    {
        Left,
        Right,
        Middle,
    }

    /// <summary>
    /// InputManager handles the querying and processing of input
    /// data from the mouse and keyboard, and provides functions
    /// for retrieving the current input state as well as its
    /// change in state from frame to frame. Input handling
    /// callbacks are supported through action maps, which can
    /// be pushed and popped from the input manager's internal
    /// stack
    /// </summary>
    public class InputManager : GameComponent, IInputManager
    {
        /// <summary>
        /// Initializes the input manager, adding an InputManager
        /// instance to the specified game's component set and
        /// adding the IInputManager interface to the game's
        /// services
        /// </summary>
        /// <param name="game">The game using the input manager</param>
        public static void Initialize(Game game)
        {
            InputManager mgr = new InputManager(game);

            game.Components.Add(mgr);
            game.Services.AddService(typeof(IInputManager), mgr);
        }

        /// <summary>
        /// Defines the state of an input set, either current
        /// or previous
        /// </summary>
        private enum InputState
        {
            Current,
            Previous
        }

        private Dictionary<InputState, XI.KeyboardState>    mKeyboard;
        private Dictionary<InputState, XI.MouseState>       mMouse;

        private Stack<ActionMap>                            mActionMaps;

        /// <summary>
        /// Gets the current keyboard state
        /// </summary>
        public XI.KeyboardState Keyboard
        {
            get { return mKeyboard[InputState.Current]; }
        }

        /// <summary>
        /// Gets the previous keyboard state (shorthand accessor)
        /// </summary>
        private XI.KeyboardState KeyboardPrevious
        {
            get { return mKeyboard[InputState.Previous]; }
        }

        /// <summary>
        /// Gets the current mouse state
        /// </summary>
        public XI.MouseState Mouse
        {
            get { return mMouse[InputState.Current]; }
        }

        /// <summary>
        /// Private shorthand accessor
        /// </summary>
        private XI.MouseState MousePrevious
        {
            get { return mMouse[InputState.Previous]; }
        }

        /// <summary>
        /// Gets the change in the mouse x position since input
        /// was last updated
        /// </summary>
        public int DX
        {
            get { return Mouse.X - MousePrevious.X; }
        }

        /// <summary>
        /// Gets the change in the mouse y position since input
        /// was last updated
        /// </summary>
        public int DY
        {
            get { return Mouse.Y - MousePrevious.Y; }
        }

        /// <summary>
        /// Gets the change in the mouse scroll wheel position 
        /// since input was last updated
        /// </summary>
        public int DZ
        {
            get { return Mouse.ScrollWheelValue - MousePrevious.ScrollWheelValue; }
        }

        /// <summary>
        /// Gets the x position of the mouse
        /// </summary>
        public int X
        {
            get { return Mouse.X; }
        }

        /// <summary>
        /// Gets the y position of the mouse
        /// </summary>
        public int Y
        {
            get { return Mouse.Y; }
        }

        /// <summary>
        /// Gets the current value of the mouse scroll wheel
        /// </summary>
        public int Z
        {
            get { return Mouse.ScrollWheelValue; }
        }

        /// <summary>
        /// Gets the active ActionMap
        /// </summary>
        public ActionMap ActionMap
        {
            get { return mActionMaps.Peek(); }
        }

        /// <summary>
        /// Returns true if the specified keyboard key is pressed
        /// </summary>
        /// <param name="key">The key to test for a key pressed state</param>
        /// <returns>True if the specified keyboard key is pressed, false otherwise</returns>
        public bool this[XI.Keys key]
        {
            get { return (Keyboard[key] == XI.KeyState.Down); }
        }

        /// <summary>
        /// Returns true if the specified mouse button is pressed
        /// </summary>
        /// <param name="button">The mouse button to test for a pressed state</param>
        /// <returns>True if the specified mouse button is pressed</returns>
        public bool this[MouseButtons button]
        {
            get
            {
                switch (button)
                {
                    case MouseButtons.Left:
                        return (Mouse.LeftButton == XI.ButtonState.Pressed);

                    case MouseButtons.Right:
                        return (Mouse.RightButton == XI.ButtonState.Pressed);

                    case MouseButtons.Middle:
                        return (Mouse.MiddleButton == XI.ButtonState.Pressed);
                }

                return false;
            }
        }

        /// <summary>
        /// Creates a new instance of Input
        /// </summary>
        /// <param name="game">The XNA Game instance</param>
        public InputManager(Game game) 
            : base(game)
        {
            mKeyboard = new Dictionary<InputState, XI.KeyboardState>();
            mMouse = new Dictionary<InputState, XI.MouseState>();

            mKeyboard[InputState.Current] =
                mKeyboard[InputState.Previous] = XI.Keyboard.GetState();

            mMouse[InputState.Current] =
                mMouse[InputState.Previous] = XI.Mouse.GetState();

            mActionMaps = new Stack<ActionMap>();
            mActionMaps.Push(new ActionMap());
        }

        /// <summary>
        /// Pushes an ActionMap onto the action map stack
        /// </summary>
        /// <param name="map">The ActionMap to push onto the ActionMap stack</param>
        public void PushActionMap(ActionMap map)
        {
            mActionMaps.Push(map);
        }

        /// <summary>
        /// Pops the topmost ActionMap from the ActionMap stack
        /// </summary>
        /// <returns>The ActionMap that was popped from the ActionMap stack</returns>
        public ActionMap PopActionMap()
        {
            if (mActionMaps.Count <= 1)
                throw new Exception("Attempt to pop the default ActionMap from the InputManager's ActionMap stack");

            return mActionMaps.Pop();
        }

        /// <summary>
        /// Updates the keyboard and mouse states
        /// </summary>
        /// <param name="gameTime">Current XNA game time</param>
        public override void Update(GameTime gameTime)
        {
            mKeyboard[InputState.Previous] = mKeyboard[InputState.Current];
            mMouse[InputState.Previous] = mMouse[InputState.Current];

            mKeyboard[InputState.Current] = XI.Keyboard.GetState();
            mMouse[InputState.Current] = XI.Mouse.GetState();

            ActionMap map = mActionMaps.Peek();

            map.ProcessInput(mKeyboard[InputState.Current], mKeyboard[InputState.Previous], 
                mMouse[InputState.Current], mMouse[InputState.Previous], gameTime);

            base.Update(gameTime);
        }

        /// <summary>
        /// Returns true if the specified keyboard key is pressed
        /// </summary>
        /// <param name="key">The key to test for a key pressed state</param>
        /// <returns>True if the specified keyboard key is pressed, false otherwise</returns>
        public bool Key(XI.Keys key)
        {
            return this[key];
        }

        /// <summary>
        /// Returns true if a key down event occurred for the specified key
        /// </summary>
        /// <param name="key">The key to test for a key down event</param>
        /// <returns>True if the specified keyboard key triggered a key down
        /// event, false otherwise</returns>
        public bool KeyDown(XI.Keys key)
        {
            return (this[key] && !this[key]);
        }

        /// <summary>
        /// Returns true if a key up event occurred for the specified key
        /// </summary>
        /// <param name="key">The key to test for a key up event</param>
        /// <returns>True if the specified keyboard key triggered a key up
        /// event, false otherwise</returns>
        public bool KeyUp(XI.Keys key)
        {
            return (!this[key] && this[key]);
        }

        /// <summary>
        /// Returns true if the specified mouse button is pressed
        /// </summary>
        /// <param name="button">The mouse button to test for a button pressed state</param>
        /// <returns>True if the specified mouse button is pressed, false otherwise</returns>
        public bool MouseButton(MouseButtons button)
        {
            return this[button];
        }

        /// <summary>
        /// Returns true if a mouse down event occurred for the specified
        /// mouse button
        /// </summary>
        /// <param name="button">The mouse button to test for a mouse down
        /// event</param>
        /// <returns>True if the specified mouse button triggered a mouse
        /// down event</returns>
        public bool MouseDown(MouseButtons button)
        {
            return (this[button] && !this[button]);
        }

        /// <summary>
        /// Returns true if a mouse up event occurred for the specified
        /// mouse button
        /// </summary>
        /// <param name="button">The mouse button to test for a mouse up
        /// event</param>
        /// <returns>True if the specified mouse button triggered a mouse
        /// up event
        /// </returns>
        public bool MouseUp(MouseButtons button)
        {
            return (!this[button] && this[button]);
        }
    }
}